#!/usr/bin/env python
# coding: UTF-8

# Copyright (C) 2015 Markus Schmidt <schmidt@boomshop.net>
# License: WTFPL - have a look in the COPYING file that came with this
# software

import sys
import os
import shutil
import re
import subprocess
import string
import signal
import gobject
import gtk
import math
import cairo
from threading import Thread
from threading import Timer
from commands import getoutput
from ast import literal_eval

try:
     import pygtk
     pygtk.require("2.6")
except:
      pass

gobject.threads_init()

basedir = os.path.dirname(os.path.realpath(__file__))
if not os.path.exists(os.path.join(basedir, "recjack.py")):
    if os.path.exists(os.path.join(os.getcwd(), "recjack.py")):
        basedir = os.getcwd()
sys.path.insert(0, basedir)

class RecJack:
    _VERSION = open(os.path.join(sys.path[0], 'version')).read().strip()
    _TITLE = "RecJack " + _VERSION
    
    iconSize       = 40
    displayHeight  = 40
    displayPadding = 2
    
    meterL         = 0
    meterR         = 0
    holdL          = 0
    holdR          = 0
    
    softL          = 0
    softR          = 0
    soften         = 2
    
    led_size       = 1
    
    recorder       = None
    state          = 0
    RBstate        = False
    RBtime         = 0.66
    RBtimer        = None
    
    RDtimer        = None
    RDtime         = 0.04
    
    background     = None
    
    regex_L        = r"^00\:\|([\*\- ]*)\|"
    regex_R        = r"^01\:\|([\*\- ]*)\|"
    
    last_file      = ""
    win_size       = [0, 0]
    
    config         = {
        "directory" : "",
        "prefix"    : "jack_capture_",
        "format"    : "",
        "digits"    : 1,
        "bits"      : "float",
        "channels"  : 2,
        "duration"  : 0,
        "sync"      : 0,
        "advanced"  : "",
        "geometry"  : None
    }
    
    bits           = [ "float", "8", "16", "24", "32" ]
    
    conf_file      = os.path.expanduser('~/.recjackrc')
    
    def __init__ (self):
        # configuration
        try:
            fconf = dict(line.strip().split('=', 1) for line in open(self.conf_file))
        except:
            fconf = {}
        
        for i in fconf:
            if not i in self.config:
                continue
            if isinstance(self.config[i], int):
                try:
                    self.config[i] = int(fconf[i])
                except: 1
            elif isinstance(self.config[i], float):
                try:
                    self.config[i] = float(fconf[i])
                except: 1
            else:
                self.config[i] = fconf[i]
        # build window
        self.create_window()
        if self.config['geometry']:
            (g, w, h, x, y) = literal_eval(self.config['geometry'])
            print g, w, h, x, y
            self.window.parse_geometry("%dx%d+%d+%d" % (w, h, x, y))
            self.window.resize(w, h)
        os.chdir(self.filechooser.get_filename())
            
    ####################
    # USER INTERACTION #
    ####################
    
    def start_clicked (self, widget):
        if self.state == 0:
            # start recording
            com = self.create_command()
            self.recorder = Recorder(com, self.recorderCB)
            print "\n### Starting jack_capture ###"
            print com + "\n"
        else:
            # stop recording
            if self.recorder:
                self.recorder.stop()
        if self.jack_check.get_active() and self.state != -1:
            self.set_UI_state(-1)
        else:
            self.start.set_sensitive(False)
            
    def channel_changed (self, widget):
        if widget.get_value() > 2:
            self.format_select.set_sensitive(False)
        else:
            self.format_select.set_sensitive(True)
            
    def resized (self, widget):
        self.display.queue_draw()
    
    def filechooser_changed(self, widget):
        os.chdir(widget.get_filename())
        
    ########################
    # RECORDER INTERACTION #
    ########################
    
    def create_command(self):
        com = "jack_capture"
        # prefix
        if self.file_prefix.get_text() and self.digit_spin.get_value() >= 0:
            com += " -fp \"%s\"" % self.file_prefix.get_text()
        # digits
        if self.digit_spin.get_value() >= 0 and self.digit_spin.get_value() != 1:
            com += " -z %d" % self.digit_spin.get_value()
        # sync to jack
        if self.jack_check.get_active():
            com += " -jt"
        # duration
        if self.time_spin.get_value() > 0:
            com += " -d %d" % self.time_spin.get_value()
        # channels
        if self.time_spin.get_value() != 2:
            com += " -c %d" % self.channel_spin.get_value()
        # format
        if self.format_select.get_active() > 0 and self.channel_spin.get_value() <= 2:
            com += " -f %s" % self.format_select.get_active_text()
        # bits
        if self.bit_select.get_active_text() != "float":
            com += " -b %s" % self.bit_select.get_active_text()
        # advanced
        if self.additional.get_text():
            com += " %s" % self.additional.get_text()
        # fixed file name
        if self.file_prefix.get_text() and self.digit_spin.get_value() < 0:
            com += " \"%s\"" % os.path.join(os.getcwd(), self.file_prefix.get_text())
        return com
    
    def recorderCB (self, com, state):
        if not state:
            # recording done
            self.set_UI_state(0)
        else:
            # strip ANSI
            com = re.sub(r'\x1b\[([0-9,A-Z]{1,2}(;[0-9]{1,2})?(;[0-9]{3})?)?[m|K]?', '', com)
            com = com.strip(string.whitespace)
            if not com:
                return
            if '""""""' in com:
                return
            # left meter
            m = re.search(self.regex_L, com)
            if m:
                s = m.group(1)
                l = len(s)
                self.led_size = 1. / l
                v = float(s.count("-")) / l
                if v > self.meterL:
                    self.meterL = v
                    self.softL = self.soften
                try:
                    self.holdL = float(s.index("*")) / l
                except:
                    1
                #self.display.queue_draw()
                if not self.RDtimer:
                    self.RDtimer = Timer(self.RDtime, self.RDfun)
                    self.RDtimer.start()
                return
            # right meter
            m = re.search(self.regex_R, com)
            if m:
                s = m.group(1)
                l = len(s)
                self.led_size = 1. / l
                v = float(s.count("-")) / l
                if v > self.meterR:
                    self.meterR = v
                    self.softR = self.soften
                try:
                    self.holdR = float(s.index("*")) / l
                except:
                    1
                #self.display.queue_draw()
                if not self.RDtimer:
                    self.RDtimer = Timer(self.RDtime, self.RDfun)
                    self.RDtimer.start()
                return
                
            if com[:17] == ">>> Recording to ":
                m = re.search(r'\"(.*)\"', com)
                self.last_file = m.group(1)
            
            if self.state == -1 and "Received JackTranportRolling. Recording." in com:
                self.set_UI_state(1)
                return
            
            if com[:7] == "Buffer:":
                if self.state == 0:
                    self.set_UI_state(1)
            else:
                print com
            
            if "Please wait while writing all data to disk" in com:
                self.set_UI_state(-1)
                self.start.set_sensitive(False)
            elif "Finished." in com:
                self.set_UI_state(0)
                
            if "Buffer:"   in com \
            or "Warning:"  in com \
            or "error"     in com \
            or "jack server is not running or cannot be started" in com \
            or "Please wait while writing" in com:
                gobject.idle_add(self.status.set_text, com)
    
    ################
    # DRAWING AREA #
    ################
    
    def draw_display(self, widget, data):
        rect = self.display.get_allocation()
        x0 = rect.x
        y0 = rect.y
        w = rect.width
        h = rect.height
        p = self.displayPadding
        _h = int((h - p * 2 - p) / 2)
        _w = w - 2 * p - self.iconSize
        _i = self.iconSize - 2 * p
        ls = int(self.led_size * _w)
        
        if w != self.win_size[0] or h != self.win_size[1] or not self.background:
            self.win_size = [w, h]
            self.draw_background()
        
        #col = self.display.get_property("style").bg[gtk.STATE_NORMAL]
        #r = col.red_float
        #g = col.green_float
        #b = col.blue_float
        r = g = b = 0.1;
        
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h)
        canvas = cairo.Context(surface)
        cr = gtk.gdk.CairoContext(canvas)
        
        # background
        cr.set_source_surface(self.background)
        cr.paint()
        
        # meters
        cr.set_source_rgba(r, g, b, 0.6)
        
        if self.holdL > self.meterL:
            # Left meter and hold
            hp = int(_w * self.holdL)
            mp = int(_w * self.meterL)
            cr.rectangle(p + hp, p, _w - hp, _h)
            cr.fill()
            if mp + ls < hp:
                cr.rectangle(p + mp, p, (hp - mp) - ls, _h)
                cr.fill()
        else:
            # Left meter
            mp = int(_w * self.meterL)
            cr.rectangle(p + mp, p, _w - mp, _h)
            cr.fill()
        
        if self.holdR > self.meterR:
            # Right meter and hold
            hp = int(_w * self.holdR)
            mp = int(_w * self.meterR)
            cr.rectangle(p + hp, p * 2 + _h, _w - hp, _h)
            cr.fill()
            if mp + ls < hp:
                cr.rectangle(p + mp, p * 2 + _h, (hp - mp) - ls, _h)
                cr.fill()
        else:
            # Right meter
            mp = int(self.meterR * _w)
            cr.rectangle(p + mp, p * 2 + _h, _w - mp, _h)
            cr.fill()
        
        # state icon
        if self.state == 1:
            cr.arc(3 * p + _w + _i / 2, p + _i / 2, _i / 4, 0, 2 * math.pi)
            cr.set_source_rgb(1 if self.RBstate else 0.33, 0, 0)
        elif self.state == -1:
            cr.set_source_rgba(0.4, 0.4, 0.4, 1 if self.RBstate else 0.33)
            cr.rectangle(3 * p + _w + _i / 4, p + _i / 4,
                         (_i / 2) * 0.4, _i / 2)
            cr.fill()
            cr.rectangle(3 * p + _w + _i / 4 + (_i / 2) * 0.6, p + _i / 4,
                         (_i / 2) * 0.4, _i / 2)
        else:
            cr.rectangle(3 * p + _w + _i / 4, p + _i / 4,
                         _i / 2, _i / 2)
            cr.set_source_rgb(0.4, 0.4, 0.4)
        cr.fill()
        
        # copy to x window
        cr = self.display.window.cairo_create()
        cr.set_source_surface(surface)
        cr.paint()
        
        # falling of meters
        #m = 0.99995
        m = ls * 0.001
        if self.softL > 0:
            self.softL -= 1
        else:
            self.meterL -= m
            
        if self.softR > 0:
            self.softR -= 1
        else:
            self.meterR -= m
            
        if self.meterL < 0.005 and self.meterR < 0.005 and self.holdL < 0.005 and self.holdR < 0.005:
           self.meterL = self.meterR = self.holdL = self.holdR = 0
        else:
            if not self.RDtimer:
                self.RDtimer = Timer(self.RDtime, self.RDfun)
                self.RDtimer.start()
            
        
    def draw_background(self):
        rect = self.display.get_allocation()
        x0 = rect.x
        y0 = rect.y
        w = rect.width
        h = rect.height
        p = self.displayPadding
        _h = int((h - p * 2 - p) / 2)
        _w = w - 2 * p - self.iconSize
        
        #col = self.display.get_property("style").bg[gtk.STATE_NORMAL]
        #r = col.red_float
        #g = col.green_float
        #b = col.blue_float
        r = g = b = 0.1;
        
        grad = cairo.LinearGradient(p, p, _w - self.iconSize, p)
        grad.add_color_stop_rgb(0.0,   0.0,  0.2, 0.75)
        grad.add_color_stop_rgb(0.5,   0.0,  0.8, 0.8)
        grad.add_color_stop_rgb(0.75,  0.25, 0.8, 0.2)
        grad.add_color_stop_rgb(0.875, 1.0,  1.0, 0.0)
        grad.add_color_stop_rgb(0.94,  1.0,  0.3, 0.0)
        
        self.background = cairo.ImageSurface(cairo.FORMAT_ARGB32, w, h)
        canvas = cairo.Context(self.background)
        cr = gtk.gdk.CairoContext(canvas)
        
        # background color
        cr.rectangle(0, 0, w, h)
        cr.set_source_rgb(r, g, b)
        cr.fill()
        
        # bar background gradients
        cr.set_line_width(2)
        cr.rectangle(p, p, _w, _h)
        cr.set_source(grad)
        cr.fill()
        cr.set_source_rgb(0, 0, 0)
        cr.stroke()
        cr.rectangle(p, 2 * p + _h, _w, _h)
        cr.set_source(grad)
        cr.fill()
        cr.set_source_rgb(0, 0, 0)
        cr.stroke()
        
        # led separators
        cr.set_source_rgba(r, g, b, 0.33)
        i = p + 3
        while (i < p + _w):
            cr.rectangle(i, p, 1, _h * 2 + p)
            cr.fill()
            i += 3
        
        # blinder
        s = float(p) / _h / 4.
        grad = cairo.LinearGradient(p, p, p, _h * 2 + p * 2)
        grad.add_color_stop_rgba(0.0,  r, g, b, 0.25)
        grad.add_color_stop_rgba(0.25 - s / 2., r, g, b, 0.0)
        grad.add_color_stop_rgba(0.5 - s,  r, g, b, 0.25)
        grad.add_color_stop_rgba(0.5 + s,  r, g, b, 0.25)
        grad.add_color_stop_rgba(0.75 + s / 2., r, g, b, 0.0)
        grad.add_color_stop_rgba(1.0,  r, g, b, 0.25)
        cr.set_source(grad)
        cr.rectangle(p, p, _w, 2 * _h + p)
        cr.fill()
    
    ##############    
    # UI HELPERS #
    ##############
    
    def set_UI_state (self, state):
        self.start.set_sensitive(True)
        self.state = state
        if state != 0:
            self.status.set_text("")
            if self.RBtimer:
                self.RBtimer.cancel()
            self.RBstate = True
            self.RBtimer = Timer(self.RBtime, self.RBfun)
            self.RBtimer.start()
            self.start.set_label(gtk.STOCK_MEDIA_STOP)
        else:
            if self.RBtimer:
                self.RBtimer.cancel()
            self.start.set_label(gtk.STOCK_MEDIA_RECORD)
            self.meterR = self.meterL = self.holdL = self.holdR = 0
            if self.last_file:
                gobject.idle_add(self.status.set_text, "Recorded to: %s" % os.path.join(os.getcwd(), self.last_file))
                self.last_file = ""
        self.display.queue_draw()
        
    def RDfun (self):
        self.RDtimer = None
        self.display.queue_draw()
        
    def RBfun (self):
        self.RBstate = not self.RBstate
        if self.state != 0:
            self.RBtimer = Timer(self.RBtime, self.RBfun)
            self.RBtimer.start()
        self.display.queue_draw()
    
    ##########
    # WINDOW #
    ##########
    
    def create_window (self):
        # WINDOW
        self.window = gtk.Window()
        self.window.set_title(self._TITLE)
        self.window.connect("delete-event", self.destroy)
        self.window.connect('check-resize', self.resized)
        self.window.set_icon_from_file(os.path.join(sys.path[0], "recjack.svg"))
        
        # MAIN ELEMENTS
        self.vbox = gtk.VBox()
        
        self.table = gtk.Table(3, 4)
        self.table.set_col_spacings(8)
        self.table.set_row_spacings(8)
        self.table.set_border_width(8)
        
        # LABELS
        self.label_prefix = gtk.Label("Directory")
        self.label_prefix.set_tooltip_text("The folder the recording is stored to.")
        self.label_prefix.set_alignment(1, 0.5)
        self.table.attach(self.label_prefix, 0, 1, 0, 1)
        
        self.label_prefix = gtk.Label("File Prefix")
        self.label_prefix.set_tooltip_text("The file name is generated from this entry, an ascending number with leading zeros as set under \"Digits\" and a suffix according to the \"Format\" selection. If \"Digits\" is set to -1, this entry is taken as the file name. You have to add a suffix by hand then.")
        self.label_prefix.set_alignment(1, 0.5)
        self.table.attach(self.label_prefix, 0, 1, 1, 2)
        
        self.label_format = gtk.Label("File Format")
        self.label_format.set_tooltip_text("Choose in which format the file is saved.")
        self.label_format.set_alignment(1, 0.5)
        self.table.attach(self.label_format, 0, 1, 2, 3)
        
        self.label_digits = gtk.Label("Digits")
        self.label_digits.set_tooltip_text("Amount of digits for the automatic file name generation. If set to -1 the file name is the entry under \"File Prefix\".")
        self.label_digits.set_alignment(1, 0.5)
        self.table.attach(self.label_digits, 2, 3, 2, 3)
        
        self.label_bits = gtk.Label("Bit Depth")
        self.label_bits.set_tooltip_text("Set the bit depth of the recorded audio file.")
        self.label_bits.set_alignment(1, 0.5)
        self.table.attach(self.label_bits, 0, 1, 3, 4)
        
        self.label_channels = gtk.Label("Channels")
        self.label_channels.set_tooltip_text("Set the amount of channels to record. Channels above 2 force the format to wavex!")
        self.label_channels.set_alignment(1, 0.5)
        self.table.attach(self.label_channels, 2, 3, 3, 4)
        
        self.label_duration = gtk.Label("Duration (s)")
        self.label_duration.set_tooltip_text("Set to a value > 0 if the recording should stop after n seconds.")
        self.label_duration.set_alignment(1, 0.5)
        self.table.attach(self.label_duration, 0, 1, 5, 6)
        
        self.label_sync = gtk.Label("Jack Sync")
        self.label_sync.set_tooltip_text("If checked, the application waits for Jack transport to start and stop.")
        self.label_sync.set_alignment(1, 0.5)
        self.table.attach(self.label_sync, 2, 3, 5, 6)
        
        self.label_additional = gtk.Label("Advanced")
        self.label_additional.set_tooltip_text("Set additional options for jack_capture. This entry is appended to the jack_capture command line before an eventual fixed file name (Digits = -1).")
        self.label_additional.set_alignment(1, 0.5)
        self.table.attach(self.label_additional, 0, 1, 6, 7)
        
        # SETTINGS
        self.filechooser = gtk.FileChooserButton("Save File To...")
        self.filechooser.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
        if self.config["directory"]:
            self.filechooser.set_current_folder(self.config["directory"])
        else:
            self.filechooser.set_current_folder(os.getcwd())
        self.filechooser.connect("selection-changed", self.filechooser_changed)
        self.table.attach(self.filechooser, 1, 4, 0, 1)
        
        self.file_prefix = gtk.Entry()
        self.file_prefix.set_text(self.config["prefix"])
        self.table.attach(self.file_prefix, 1, 4, 1, 2)
        
        self.format_select = gtk.combo_box_new_text()
        self.format_select.append_text("Default")
        output = subprocess.check_output('jack_capture -pf', shell = True)
        self.formats = [i.strip() for i in output.split(" ")]
        for i in self.formats:
            if i:
                self.format_select.append_text(i)
        if self.config["format"] and self.config["format"] in self.formats:
            self.format_select.set_active(self.formats.index(self.config['format']) + 1)
        else:
            self.format_select.set_active(0)
        self.table.attach(self.format_select, 1, 2, 2, 3, gtk.EXPAND | gtk.FILL, gtk.EXPAND)
        
        self.digit_adj = gtk.Adjustment(value=-1, lower=-1, upper=10, step_incr=1, page_incr=2, page_size=0)
        self.digit_spin = gtk.SpinButton()
        self.digit_spin.set_adjustment(self.digit_adj)
        self.digit_spin.set_value(self.config["digits"])
        self.table.attach(self.digit_spin, 3, 4, 2, 3)
        
        self.bit_select = gtk.combo_box_new_text()
        for i in self.bits:
            self.bit_select.append_text(str(i))
        if self.config['bits'] in self.bits:
            self.bit_select.set_active(self.bits.index(self.config['bits']))
        else:
            self.bit_select.set_active(0)
        self.table.attach(self.bit_select, 1, 2, 3, 4, gtk.EXPAND | gtk.FILL, gtk.EXPAND)
        
        self.channel_adj = gtk.Adjustment(value=2, lower=1, upper=32, step_incr=1, page_incr=2, page_size=0)
        self.channel_spin = gtk.SpinButton()
        self.channel_spin.set_adjustment(self.channel_adj)
        self.channel_spin.set_value(self.config["channels"])
        self.table.attach(self.channel_spin, 3, 4, 3, 4)
        self.channel_spin.connect("value-changed", self.channel_changed)
        
        self.time_adj = gtk.Adjustment(value=-1, lower=0, upper=100000000, step_incr=1, page_incr=10, page_size=0)
        self.time_spin = gtk.SpinButton()
        self.time_spin.set_adjustment(self.time_adj)
        self.time_spin.set_value(self.config["duration"])
        self.table.attach(self.time_spin, 1, 2, 5, 6)
        
        self.jack_check = gtk.CheckButton()
        self.jack_check.set_active(self.config["sync"])
        self.table.attach(self.jack_check, 3, 4, 5, 6)
        
        self.additional = gtk.Entry()
        self.additional.set_text(self.config["advanced"])
        self.table.attach(self.additional, 1, 4, 6, 7)
        
        self.table.attach(gtk.HSeparator(), 0, 4, 4, 5)
        
        # BUTTONS
        self.buttons = gtk.HBox(True, 8)
        self.buttons.set_border_width(8)
        
        self.start = gtk.Button(stock = gtk.STOCK_MEDIA_RECORD)
        self.start.connect("clicked", self.start_clicked)
        self.buttons.pack_start(self.start)
        self.start.set_size_request(-1, 48)
        
        # DRAWING AREA
        self.display = gtk.DrawingArea()
        self.display.connect("expose_event", self.draw_display)
        self.display.set_size_request(-1, self.displayHeight)
        
        # STATUS BAR
        self.status = gtk.Label()
        
        # STITCH TOGETHER
        self.window.add(self.vbox)
        self.vbox.pack_start(self.table, True, True)
        self.vbox.pack_start(self.buttons, True, True)
        self.vbox.pack_start(self.display, False, True)
        self.vbox.pack_end(self.status, expand=False)
        
        self.window.show_all()
        
        if not self.background:
            self.resized(None)
            
        # Ctrl+C
        signal.signal(signal.SIGINT, self.ctrlC)
        signal.signal(signal.SIGHUP, self.ctrlC)
        signal.signal(signal.SIGQUIT, self.ctrlC)
        signal.signal(signal.SIGABRT, self.ctrlC)
        signal.signal(signal.SIGSEGV, self.ctrlC)
        signal.signal(signal.SIGTERM, self.ctrlC)
        
    def ctrlC(self, signum, frame):
        self.destroy(None, None)
        
    def destroy (self, widget, data):
        if self.RBtimer:
            self.RBtimer.cancel()
        if self.recorder:
            self.recorder.stop()
            
        # configuration
        w = self.window
        conf = {
            "directory" : self.filechooser.get_filename(),
            "prefix"    : self.file_prefix.get_text(),
            "digits"    : int(self.digit_spin.get_value()),
            "duration"  : int(self.time_spin.get_value()),
            "channels"  : int(self.channel_spin.get_value()),
            "sync"      : 1 if self.jack_check.get_active() else 0,
            "format"    : self.format_select.get_active_text(),
            "bits"      : self.bit_select.get_active_text(),
            "advanced"  : self.additional.get_text(),
            "geometry"  : str((int(w.get_gravity()),) +  w.get_size() + w.get_position())
        }
        with open(self.conf_file, "w") as f:
            for i in conf:
                f.write(i + "=" + str(conf[i]) + "\n")
        gtk.main_quit()

class Recorder (Thread):
    def __init__ (self, command, callback):
        Thread.__init__(self)
        self.daemon   = True
        self.command  = command
        self.callback = callback
        self.start()
        
    def run (self):
        self.recording = True
        self.process = subprocess.Popen(self.command,
            stdout = subprocess.PIPE,
            stderr = subprocess.STDOUT,
            stdin  = subprocess.PIPE,
            universal_newlines = True,
            shell = True,
            preexec_fn=os.setsid)
        
        lastline = ""
        while self.process.poll() == None and self.recording:
            out = self.process.stdout.readline().strip()
            self.callback(out, True)
            #print out.strip()
        
        self.process.stdout.close()
        self.process.stdin.close()
        self.recording = False
        self.process = None
        self.callback("", False)
        
    def stop (self):
        if self.recording and self.process:
            self.process.stdin.write("\n")
            self.recording = False
        
if __name__ =="__main__":
    recjack = RecJack()
    gtk.main()
